@@ -27,11 +27,11 @@ class RunScriptAssistant : public Cantor::Assistant
 {
   public:
     RunScriptAssistant( QObject* parent, QList<QVariant> args );
-    ~RunScriptAssistant() override;
+    ~RunScriptAssistant() override = default;
 
-    void initActions() Q_DECL_OVERRIDE;
+    void initActions() override;
     
-    QStringList run(QWidget* parentt) Q_DECL_OVERRIDE;
+    QStringList run(QWidget* parentt) override;
     
 };
 
